use core::str;
use std::collections::HashMap;

use super::*;
use crate::ngen_toml::*;

#[derive(Debug, Clone)]
pub struct Target<'a> {
    pub outfile: &'a str,
    pub compiler: &'a str,
    pub compiler_flags: Vec<&'a str>,
    pub linker: &'a str,
    pub linker_flags: Vec<&'a str>,
    pub linker_libs: Vec<&'a str>,
    pub sources: Vec<&'a str>,
    pub opts: TargetOptions<'a>,
}

impl<'a> Target<'a> {
    pub fn get_default(
        ) -> Self {
        Target {
            outfile: "a.out",
            compiler: "cc",
            compiler_flags: vec![],
            linker: "cc",
            linker_flags: vec![],
            linker_libs: vec![],
            sources: vec![],
            opts: TargetOptions::get_default(),
        }
    }
    pub fn from_parsed_target(
        parsed_target: &'a ParsedTarget,
        name: &'a str,
        all_parsed_targets: &'a HashMap<String, ParsedTarget>,
        all_targets: &HashMap<&'a str, Target<'a>>,
        inheritance_chain: &mut HashMap<&'a str, &'a str>
    ) -> Result<HashMap<&'a str, Target<'a>>, ConfigurationError> {
        let mut retmap: HashMap<&str, Target> = HashMap::new();
        let mut ret = Self::get_default();

        if ret.opts.inherit_from == name {
            ret.opts.inherit = false;
        }

        if let Some(v) = &parsed_target.opts {
            ret.opts = TargetOptions::from_parsed_target_options(v)?;
        }

        if ret.opts.inherit {
            let inherit_target: Target;

            match all_targets.get(ret.opts.inherit_from) {
                Some(v) => {
                    inherit_target = v.clone()
                }
                None => {
                    match all_parsed_targets.get(ret.opts.inherit_from) {
                        Some(v) => {
                            let inherited_targets: HashMap<&str, Target>;
                            if let Some(_) = inheritance_chain.get(ret.opts.inherit_from) {
                                let mut message = format!("inheritance dependency cylce detected: ");
                                for (targ_name, depends_on) in inheritance_chain {
                                    message.push_str(&format!("`{}` depends on `{}`, ", targ_name, depends_on));
                                }
                                message.push_str(&format!("and `{}` depends on `{}`", name, ret.opts.inherit_from));
                                return Err(ConfigurationError { message });
                            }
                            inheritance_chain.insert(name, ret.opts.inherit_from);
                            inherited_targets = Self::from_parsed_target(v, ret.opts.inherit_from, all_parsed_targets, all_targets, inheritance_chain)?;
                            inherit_target = inherited_targets.get(ret.opts.inherit_from).unwrap().clone();
                            retmap.extend(inherited_targets.into_iter());
                        }
                        None => {
                            return Err(ConfigurationError { 
                                message: 
                                    format!("target `{0}` trying to inherit from target `{1}`, but target `{1}` does not exist", 
                                            name, 
                                            ret.opts.inherit_from)
                            });
                        }
                    }
                }
            }


            if let None = parsed_target.outfile {
                ret.outfile = inherit_target.outfile;
            }
            if let None = parsed_target.compiler {
                ret.compiler = inherit_target.compiler;
            }
            if let None = parsed_target.linker {
                ret.linker = inherit_target.linker;
            }

            for string in inherit_target.compiler_flags {
                ret.compiler_flags.push(string);
            }
            for string in inherit_target.linker_flags {
                ret.linker_flags.push(string);
            }
            for string in inherit_target.linker_libs {
                ret.linker_libs.push(string);
            }
            for string in inherit_target.sources {
                ret.sources.push(string);
            }
        }

        if let Some(v) = &parsed_target.outfile {
            ret.outfile = v;
        }
        if let Some(v) = &parsed_target.compiler {
            ret.compiler = v;
        }
        if let Some(v) = &parsed_target.compiler_flags {
            for string in v {
                ret.compiler_flags.push(string);
            }
        }
        if let Some(v) = &parsed_target.linker {
            ret.linker = v;
        } else {
            if let Some(v) = &parsed_target.compiler {
                ret.linker = v;
            }
        }
        if let Some(v) = &parsed_target.linker_flags {
            for string in v {
                ret.linker_flags.push(string);
            }
        }
        if let Some(v) = &parsed_target.linker_libs {
            for string in v {
                ret.linker_libs.push(string);
            }
        }
        if let Some(v) = &parsed_target.sources {
            for string in v {
                ret.sources.push(string);
            }
        }

        retmap.insert(name, ret);

        Ok(retmap)
    }
}
